// werapitest
//
// Exercises new Vista Windows Error Reporting API
// Version 2 - now using WerReportSetParameter()
//
// Claus Brod, http://www.clausbrod.de/Blog

#define UNICODE
#define _UNICODE

#include <stdio.h>
#include <tchar.h>

#define _WIN32_WINNT 0x0501
#include <windows.h>
#include "werapi.h"

#include <atlbase.h>

#include <dbghelp.h>
#pragma comment(lib, "dbghelp.lib")
#pragma comment(lib, "version.lib")

typedef HRESULT 
(WINAPI *pfn_WERREPORTCREATE)(PCWSTR, WER_REPORT_TYPE, 
                              PWER_REPORT_INFORMATION, HREPORT *);
static pfn_WERREPORTCREATE pWerReportCreate = 0;

typedef HRESULT 
(WINAPI *pfn_WERREPORTSUBMIT)(HREPORT, WER_CONSENT, 
                              DWORD, PWER_SUBMIT_RESULT);
static pfn_WERREPORTSUBMIT pWerReportSubmit = 0;

typedef HRESULT (WINAPI *pfn_WERREPORTCLOSEHANDLE)(HREPORT);
static pfn_WERREPORTCLOSEHANDLE pWerReportCloseHandle = 0;

typedef HRESULT 
(WINAPI *pfn_WERREPORTADDDUMP)(HREPORT, HANDLE, HANDLE, 
                               WER_DUMP_TYPE, PWER_EXCEPTION_INFORMATION, 
                               PWER_DUMP_CUSTOM_OPTIONS, DWORD);
static pfn_WERREPORTADDDUMP pWerReportAddDump = 0;

typedef HRESULT 
(WINAPI *pfn_WERREPORTSETPARAMETER)(HREPORT, DWORD,
                                    PCWSTR, PCWSTR);
static pfn_WERREPORTSETPARAMETER pWerReportSetParameter = 0;

//--------- Launch debugger ----------------------------------------------------

static TCHAR *installed_debugger()
{
  CRegKey key;
  if (ERROR_SUCCESS != key.Open(HKEY_LOCAL_MACHINE, 
    __T("Software\\Microsoft\\Windows NT\\CurrentVersion\\AeDebug\\"), KEY_READ))
    return NULL;

  static TCHAR debuggerPath[1024];
  debuggerPath[0] = 0;
  DWORD dwCount = _countof(debuggerPath);
  key.QueryStringValue(__T("Debugger"), debuggerPath, &dwCount);
  return debuggerPath[0] ? debuggerPath : NULL;
}

static bool launch_debugger(void)
{
  TCHAR *debugger = installed_debugger();
  if (!debugger)
    return false;

  bool debuggerLaunched = false;

  // create event
  SECURITY_ATTRIBUTES attributes = { sizeof(attributes), 0, TRUE };
  HANDLE debugEvent = CreateEvent(&attributes, TRUE, FALSE, NULL);
  if (debugEvent) {
    // get process ID
    DWORD pid = GetCurrentProcessId();

    // build command string
    TCHAR commandBuffer[1024];
    _sntprintf_s(commandBuffer, _countof(commandBuffer), _TRUNCATE,
      debugger, pid, debugEvent);

    // start JIT debugger
    STARTUPINFO sinfo;
    memset(&sinfo, 0, sizeof(sinfo));
    sinfo.cb = sizeof(sinfo);

    PROCESS_INFORMATION pinfo;
    BOOL createProcRet = CreateProcess(NULL, commandBuffer, 
      NULL, NULL, 
      TRUE,  // inherit (event) handles
      0, NULL, NULL, &sinfo, &pinfo);

    if (createProcRet) {
      // wait until debugger fires event
      DWORD dwWaitResult = 0;
      dwWaitResult = WaitForSingleObject(debugEvent, 1000*100);
      if (dwWaitResult == WAIT_OBJECT_0) {
        debuggerLaunched = true;
      }
    } else {
      fprintf(stderr, "Could not run debugger process.\n");
    }
  }

  return debuggerLaunched;
}


//--------- WER helper functions -----------------------------------------------

static bool get_module_file_version(const TCHAR *moduleName,
                                    TCHAR *versionBuffer, size_t bufsize)
{
  _sntprintf_s(versionBuffer, bufsize, _TRUNCATE, __T("0.0.0.0"));

  DWORD dwDummy = 0;
  DWORD dwInfoSize = ::GetFileVersionInfoSize(moduleName, &dwDummy);
  if (!dwInfoSize) {
    // no version info attached
    return false;
  }

  bool ret = false;

  char *pcInfoBuffer = new char[dwInfoSize];
  if (::GetFileVersionInfo(moduleName, 0, dwInfoSize, pcInfoBuffer)) {
    VS_FIXEDFILEINFO *pFixedFile;
    UINT wInfoLen;
    if (::VerQueryValue(pcInfoBuffer, __T("\\"), (void**)&pFixedFile, &wInfoLen)) {
      _sntprintf_s(versionBuffer, bufsize, _TRUNCATE,
        __T("%d.%d.%d.%d"),
        HIWORD(pFixedFile->dwFileVersionMS),
        LOWORD(pFixedFile->dwFileVersionMS),
        HIWORD(pFixedFile->dwFileVersionLS),
        LOWORD(pFixedFile->dwFileVersionLS));
      ret = true;
    } else {
      fprintf(stderr, "VerQueryValue() fails\n");
    }
  } else {
    fprintf(stderr, "GetFileVersionInfo() fails\n");
  }
  
  delete [] pcInfoBuffer;
  return ret;
}

static void get_module_name(HMODULE hModule, TCHAR *moduleName, 
                            size_t bufsize)
{
  ::GetModuleFileName(hModule, moduleName, bufsize);
  PathStripPath(moduleName);
}


//--------- Report crash using WER APIs ----------------------------------------

// create the report handle
static HREPORT wer_report_create()
{
  WER_REPORT_INFORMATION werReportInfo;
  memset(&werReportInfo, 0, sizeof(werReportInfo));
  werReportInfo.dwSize = sizeof(werReportInfo);
  wcscpy_s(werReportInfo.wzFriendlyEventName, 
    _countof(werReportInfo.wzFriendlyEventName), L"Stopped working"); 
  wcscpy_s(werReportInfo.wzApplicationName, 
    _countof(werReportInfo.wzApplicationName), L"werapitest.exe");

  PCWSTR eventType = APPCRASH_EVENT;
  HREPORT hReportHandle;
  if (FAILED(pWerReportCreate(eventType, WerReportApplicationCrash,
    &werReportInfo, &hReportHandle)) || !hReportHandle) {
      return NULL;
  }

  return hReportHandle;
}

// Set signature parameters
static void wer_report_set_parameters(HREPORT hReportHandle, 
                                      EXCEPTION_POINTERS *exc_ptr)
{
  TCHAR moduleName[1024];
  get_module_name(NULL, moduleName, _countof(moduleName));
  pWerReportSetParameter(hReportHandle, 0, L"Application Name", moduleName);

  TCHAR buffer[1024];
  get_module_file_version(moduleName, buffer, _countof(buffer)); 
  pWerReportSetParameter(hReportHandle, 1, L"Application Version", buffer);

  HMODULE hModule = GetModuleHandle(0);
  DWORD timeStamp = GetTimestampForLoadedLibrary(hModule);
  _sntprintf_s(buffer, _countof(buffer), _TRUNCATE,
    __T("%x"), timeStamp);
  pWerReportSetParameter(hReportHandle, 2, L"Application Timestamp", buffer);

  // determine module name from crash address
  moduleName[0] = 0;
  void *exceptionAddress = exc_ptr->ExceptionRecord->ExceptionAddress;
  if (GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | 
    GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
    (LPCTSTR)exceptionAddress, &hModule)) {
    get_module_name(hModule, moduleName, _countof(moduleName));
  }
  pWerReportSetParameter(hReportHandle, 3, L"Fault Module Name", moduleName);

  get_module_file_version(moduleName, buffer, _countof(buffer)); 
  pWerReportSetParameter(hReportHandle, 4, L"Fault Module Version", buffer);

  timeStamp = GetTimestampForLoadedLibrary(hModule);
  _sntprintf_s(buffer, _countof(buffer), _TRUNCATE,__T("%x"), timeStamp);
  pWerReportSetParameter(hReportHandle, 5, L"Fault Module Timestamp", buffer);

  _sntprintf_s(buffer, _countof(buffer), _TRUNCATE,
    __T("%08x"), exc_ptr->ExceptionRecord->ExceptionCode);
  pWerReportSetParameter(hReportHandle, 6, L"Exception Code", buffer);

  INT_PTR offset = (char *)exceptionAddress - (char *)hModule;
  _sntprintf_s(buffer, _countof(buffer), _TRUNCATE, __T("%p"), offset);
  pWerReportSetParameter(hReportHandle, 7, L"Exception Offset", buffer);
}

// Add minidump to report
static bool wer_report_add_minidump(HREPORT hReportHandle,
                                    EXCEPTION_POINTERS *exc_ptr)
{
  WER_EXCEPTION_INFORMATION werExceptionInformation;
  werExceptionInformation.bClientPointers = FALSE;
  werExceptionInformation.pExceptionPointers = exc_ptr;
  bool dumpAdded = SUCCEEDED(pWerReportAddDump(hReportHandle, ::GetCurrentProcess(),
    ::GetCurrentThread(), WerDumpTypeMiniDump, &werExceptionInformation, NULL, 0));
  if (!dumpAdded) {
    fprintf(stderr, "Minidump generation failed.\n");
    return false;
  }
  return true;
}

static bool report_crash(_EXCEPTION_POINTERS *exc_ptr)
{
  // create the error report
  HREPORT hReportHandle = wer_report_create();
  if (NULL == hReportHandle) {
      fprintf(stderr, "WerReportCreate() failed.\n");
      return false;
  }

  wer_report_set_parameters(hReportHandle, exc_ptr);

  wer_report_add_minidump(hReportHandle, exc_ptr);

  // Report is complete, now submit it
  bool ret = false;
  DWORD submitOptions = WER_SUBMIT_OUTOFPROCESS | WER_SUBMIT_NO_CLOSE_UI;
  WER_SUBMIT_RESULT submitResult = (WER_SUBMIT_RESULT)0;
  HRESULT hr = pWerReportSubmit(hReportHandle, WerConsentNotAsked,
    submitOptions, &submitResult);
  if (SUCCEEDED(hr)) {
      switch(submitResult)
      {
      case WerDisabled:
        printf("Error reporting was disabled.\n");
        break;
      case WerDisabledQueue:
        printf("Queuing was disabled.\n");
        break;
      case WerReportAsync:
        printf("Report was asynchronous.\n");
        ret = true;
        break;
      case WerReportCancelled:
        printf("Report was cancelled.\n");
        break;
      case WerReportDebug:
        printf("Debug button clicked.\n");
        if (launch_debugger()) {
          printf("Debugger launched successfully.\n");
          DebugBreak();
          ret = true;
        }
        break;
      case WerReportFailed:
        printf("Report submission failed.\n");
        break;
      case WerReportQueued:
        printf("Report queued.\n");
        ret = true;
        break;
      case WerReportUploaded:
        printf("Report uploaded.\n");
        ret = true;
        break;
      default:
        printf("Unexpected return value %d from WerReportSubmit().\n",
          submitResult);
        break;
      }
  } else {
    printf("WerReportSubmit() failed\n");
  }
  pWerReportCloseHandle(hReportHandle);

  return ret;
}

static int filter_exception(EXCEPTION_POINTERS *exc_ptr)
{
  report_crash(exc_ptr);
  return EXCEPTION_EXECUTE_HANDLER;
}

static void wedding_crasher(void)
{
  __try {
    int *foo = (int *)0;
    *foo = 42;
  } __except(filter_exception(GetExceptionInformation())) {
    printf("Now in exception handler, process is still alive!\n");
  }
  Sleep(5000);
}

int main()
{
  HMODULE hWer = LoadLibrary(__T("Wer.dll"));
  if (hWer) {
    pWerReportCreate = 
      (pfn_WERREPORTCREATE)GetProcAddress(hWer, "WerReportCreate");
    pWerReportSubmit = 
      (pfn_WERREPORTSUBMIT)GetProcAddress(hWer, "WerReportSubmit");
    pWerReportCloseHandle = 
      (pfn_WERREPORTCLOSEHANDLE)GetProcAddress(hWer, "WerReportCloseHandle");
    pWerReportAddDump = 
      (pfn_WERREPORTADDDUMP)GetProcAddress(hWer, "WerReportAddDump");
    pWerReportSetParameter = 
      (pfn_WERREPORTSETPARAMETER)GetProcAddress(hWer, "WerReportSetParameter");
  }

  if (!pWerReportCreate || !pWerReportSubmit || 
    !pWerReportCloseHandle || !pWerReportAddDump || !pWerReportSetParameter) {
      printf("Cannot initialize WER API.\n");
      return 1;
  }

  wedding_crasher();
  return 0;
}
